# -*- coding: utf-8 -*-
from __future__ import print_function
"""
Module for communication with vCenter/ESX, part of virt-who

Copyright (C) 2012 Radek Novacek <rnovacek@redhat.com>

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""

import os
import suds
import suds.transport
import suds.client
import requests
import errno
import stat
import re
from six import BytesIO
import io
import logging
from time import time
from six.moves.urllib.error import URLError
import socket
from collections import defaultdict
from six.moves.http_client import HTTPException

from virtwho import virt
from virtwho.config import VirtConfigSection
from virtwho.virt import StatusReport

try:
    from urllib.parse import unquote as urldecode
except ImportError:
    from urllib import unquote as urldecode


class FileAdapter(requests.adapters.BaseAdapter):
    '''Add handler from downloading local files.

    This is necessary because we want suds to use local wsdl file.
    '''
    def send(self, request, **kwargs):
        resp = requests.Response()

        filename = request.url.replace('file://', '')
        if not os.path.isabs(filename):
            raise ValueError("Expected absolute path: %s" % request.url)

        try:
            resp.raw = io.open(filename, "rb")
            resp.raw.release_conn = resp.raw.close
        except IOError as e:
            if e.errno == errno.EACCES:
                resp.status_code = requests.codes.forbidden
            elif e.errno == errno.ENOENT:
                resp.status_code = requests.codes.not_found
            else:
                resp.status_code = requests.codes.bad_request
        else:
            resp_stat = os.fstat(resp.raw.fileno())
            if stat.S_ISREG(resp_stat.st_mode):
                resp.headers['Content-Length'] = resp_stat.st_size
            resp.status_code = requests.codes.ok
        return resp

    def close(self):
        pass


class RequestsTransport(suds.transport.Transport):
    '''
    Transport for suds that uses Requests instead of urllib2.

    This unifies network handling with other backends. For example
    proxy support will be same as for other modules.
    '''
    def __init__(self, session=None):
        suds.transport.Transport.__init__(self)
        self._session = session or requests.Session()
        self._session.mount('file://', FileAdapter())

    def open(self, request):
        resp = self._session.get(request.url, headers=request.headers, verify=False)
        resp.raise_for_status()
        return BytesIO(resp.content)

    def send(self, request):
        resp = self._session.post(
            request.url,
            data=request.message,
            headers=request.headers,
            timeout=self.options.timeout,
            verify=False
        )
        ct = resp.headers.get('content-type', '')
        if 'application/soap+xml' not in ct and 'text/xml' not in ct:
            resp.raise_for_status()
        return suds.transport.Reply(
            resp.status_code,
            resp.headers,
            resp.content,
        )


class Esx(virt.Virt):
    CONFIG_TYPE = "esx"
    MAX_WAIT_TIME = 60  # 1 minute

    def __init__(self, logger, config, dest, terminate_event=None,
                 interval=None, oneshot=False, status=False):
        super(Esx, self).__init__(logger, config, dest,
                                  terminate_event=terminate_event,
                                  interval=interval,
                                  oneshot=oneshot,
                                  status=status)
        self.url = self.config['server']
        self.username = self.config['username']
        self.password = self.config['password']

        self.filter = None
        self.sc = None

    def _prepare(self):
        """ Prepare for obtaining information from ESX server. """
        self.logger.debug("Log into ESX")
        self.login()

        self.logger.debug("Creating ESX event filter")
        self.filter = self.createFilter()

    def _cancel_wait(self):
        try:
            self.client.service.CancelWaitForUpdates(_this=self.sc.propertyCollector)
        except Exception:
            pass

    def _run(self):
        self._prepare()

        version = ''
        last_version = 'last_version'  # Bogus value so version != last_version from the start
        self.hosts = defaultdict(Host)
        self.vms = defaultdict(VM)
        self.clusters = defaultdict(Cluster)
        next_update = time()
        options = {'maxWaitSeconds': 0}

        while self._oneshot or not self.is_terminated():

            if version == '':
                # also, clean all data we have
                self.hosts.clear()
                self.vms.clear()

            try:
                # Make sure that WaitForUpdatesEx finishes even
                # if the ESX shuts down in the middle of waiting
                self.client.set_options(timeout=self.MAX_WAIT_TIME)
                self.logger.debug("calling esx service")
                updateSet = self.client.service.WaitForUpdatesEx(
                    _this=self.sc.propertyCollector,
                    version=version,
                    options=options)
            except (socket.error, URLError, requests.exceptions.Timeout):
                self.logger.debug("Wait for ESX event finished, timeout")
                self._cancel_wait()
                # Get the initial update again
                version = ''
                continue
            except (suds.WebFault, HTTPException) as e:
                suppress_exception = False
                try:
                    if hasattr(e, 'fault'):
                        if e.fault.faultstring == 'The session is not authenticated.':
                            # Do not print the exception if we get 'not authenticated',
                            # it's quite normal behaviour and nothing to worry about
                            suppress_exception = True
                        if e.fault.faultstring == 'The task was canceled by a user.':
                            # Do not print the exception if we get 'canceled by user',
                            # this happens when the wait is terminated when
                            # virt-who is being stopped
                            continue
                except Exception:
                    pass
                if not suppress_exception:
                    self.logger.exception("Waiting for ESX events fails:")
                self._cancel_wait()
                version = ''
                self._prepare()
                continue

            if self.status:
                self._send_data(data_to_send=StatusReport(self.config))
                break

            if updateSet is not None:
                version = updateSet.version
                self.applyUpdates(updateSet)

            if hasattr(updateSet, 'truncated') and updateSet.truncated:
                continue

            if last_version != version or time() > next_update:
                assoc = self.getHostGuestMapping()
                self._send_data(data_to_send=virt.HostGuestAssociationReport(self.config, assoc))
                next_update = time() + self.interval
                last_version = version

            if self._oneshot:
                break
            else:
                self.wait(self.interval)

        self.cleanup()

    def _format_hostname(self, host, domain):
        return u'{0}.{1}'.format(host, domain)

    def stop(self):
        # We need to ensure that when we stop the thread, we clean up and exit the wait
        self.cleanup()
        super(Esx, self).stop()

    def cleanup(self):
        self._cancel_wait()

        if self.filter is not None:
            try:
                self.client.service.DestroyPropertyFilter(self.filter)
            except suds.WebFault:
                pass
            self.filter = None

        self.logout()

    def getHostGuestMapping(self):
        mapping = {'hypervisors': []}
        for host_id, host in list(self.hosts.items()):
            if self.skip_for_parent(host_id, host) :
                continue
            guests = []

            try:
                if self.config['hypervisor_id'] == 'uuid':
                    uuid = host['hardware.systemInfo.uuid']
                elif self.config['hypervisor_id'] == 'hwuuid':
                    uuid = host_id
                elif self.config['hypervisor_id'] == 'hostname':
                    uuid = host['config.network.dnsConfig.hostName']
                    domain_name = host['config.network.dnsConfig.domainName']
                    if domain_name:
                        uuid = self._format_hostname(uuid, domain_name)
            except KeyError:
                self.logger.debug("Host '%s' doesn't have hypervisor_id property", host_id)
                continue

            if host['vm']:
                for vm_id in host['vm'].ManagedObjectReference:
                    if vm_id.value not in self.vms:
                        self.logger.debug("Host '%s' references non-existing guest '%s'", host_id, vm_id.value)
                        continue
                    vm = self.vms[vm_id.value]
                    if 'config.uuid' not in vm:
                        self.logger.debug("Guest '%s' doesn't have 'config.uuid' property", vm_id.value)
                        continue
                    if not vm['config.uuid'].strip():
                        self.logger.debug("Guest '%s' has empty 'config.uuid' property", vm_id.value)
                        continue
                    state = virt.Guest.STATE_UNKNOWN
                    try:
                        if vm['runtime.powerState'] == 'poweredOn':
                            state = virt.Guest.STATE_RUNNING
                        elif vm['runtime.powerState'] == 'suspended':
                            state = virt.Guest.STATE_PAUSED
                        elif vm['runtime.powerState'] == 'poweredOff':
                            state = virt.Guest.STATE_SHUTOFF
                    except KeyError:
                        self.logger.debug("Guest '%s' doesn't have 'runtime.powerState' property", vm_id.value)
                    guests.append(virt.Guest(self.getVmUuid(vm), self.CONFIG_TYPE, state))
            try:
                name = host['config.network.dnsConfig.hostName']
                domain_name = host['config.network.dnsConfig.domainName']
                if domain_name:
                    name = self._format_hostname(name, domain_name)
            except KeyError:
                self.logger.debug("Unable to determine hostname for host '%s'. Ommitting from report", uuid)
                continue

            facts = {
                virt.Hypervisor.CPU_SOCKET_FACT: str(host['hardware.cpuInfo.numCpuPackages']),
                virt.Hypervisor.HYPERVISOR_TYPE_FACT: host.get('config.product.name', 'vmware'),
                virt.Hypervisor.SYSTEM_UUID_FACT: host['hardware.systemInfo.uuid']
            }

            if host['parent'] and host['parent']._type == 'ClusterComputeResource':
                cluster_id = host['parent'].value
                # print('', self.clusters, cluster_id)
                cluster = self.clusters[cluster_id]
                facts[virt.Hypervisor.HYPERVISOR_CLUSTER] = urldecode(cluster['name'])

            version = host.get('config.product.version', None)
            if version:
                facts[virt.Hypervisor.HYPERVISOR_VERSION_FACT] = version

            mapping['hypervisors'].append(virt.Hypervisor(hypervisorId=uuid, guestIds=guests, name=name, facts=facts))
        return mapping

    def getVmUuid(self, vm):
        """
        Version >= 13 needs to be decoded using following key
        as it uses big-endian. BZ #1809098
        From: 78563412-AB90-EFCD-1234-567890ABCDEF
        To:   12345678-90AB-CDEF-1234-567890ABCDEF
        """
        s = vm['config.uuid']
        if 'config.version' not in vm:
            return s
        version = int(vm['config.version'].split('-')[1])
        if (version >= 13):
            return s[6:8] + s[4:6] + s[2:4] + s[0:2] + "-" + s[11:13] + s[9:11] + "-" + s[16:18] + s[14:16] + s[18:]
        else:
            return s

    def skip_for_parent(self, host_id, host):
        """
        Determines if the host's parent meets the criteria for inclusion/exclusion for the report
        Returns True/False based on the parent meeting the criteria
        """
        parent = host['parent'].value
        if self.config['exclude_host_parents'] is not None:
            for seg in self.config['exclude_host_parents']:
                if re.search(seg.replace("*",".*"), parent):
                    self.logger.debug("Skipping host '%s' because its parent '%s' is excluded", host_id, parent)
                    return True
        if self.config['filter_host_parents'] is not None:
            found = False
            for seg in self.config['filter_host_parents']:
                if re.search(seg.replace("*",".*"), parent):
                    found = True
            if not found:
                self.logger.debug("Skipping host '%s' because its parent '%s' is not included", host_id, parent)
                return True
        return False

    def login(self):
        """
        Log into ESX
        """

        kwargs = {'transport': RequestsTransport()}
        # Connect to the vCenter server
        if self.config['simplified_vim']:
            wsdl = 'file://%s/vimServiceMinimal.wsdl' % os.path.dirname(os.path.abspath(__file__))
            kwargs['cache'] = None
        else:
            wsdl = self.url + '/sdk/vimService.wsdl'
        try:
            self.client = suds.client.Client(wsdl, location="%s/sdk" % self.url, **kwargs)
        except requests.RequestException as e:
            raise virt.VirtError(str(e))

        self.client.set_options(timeout=self.MAX_WAIT_TIME)

        # Get Meta Object Reference to ServiceInstance which is the root object of the inventory
        self.moRef = suds.sudsobject.Property('ServiceInstance')
        self.moRef._type = 'ServiceInstance'  # pylint: disable=W0212

        # Service Content object defines properties of the ServiceInstance object
        try:
            self.sc = self.client.service.RetrieveServiceContent(_this=self.moRef)
        except requests.RequestException as e:
            raise virt.VirtError(str(e))

        # Login to server using given credentials
        try:
            # Don't log message containing password
            logging.getLogger('suds.client').setLevel(logging.CRITICAL)
            self.client.service.Login(
                _this=self.sc.sessionManager,
                userName=self.username,
                password=self.password
            )
            logging.getLogger('suds.client').setLevel(logging.ERROR)
        except requests.RequestException as e:
            raise virt.VirtError(str(e))
        except suds.WebFault as e:
            self.logger.exception("Unable to login to ESX")
            raise virt.VirtError(str(e))

    def logout(self):
        """ Log out from ESX. """
        try:
            if self.sc:
                self.client.service.Logout(_this=self.sc.sessionManager)
                self.sc = None
        except Exception as e:
            self.logger.info("Can't log out from ESX: %s", str(e))

    def createFilter(self):
        oSpec = self.objectSpec()
        oSpec.obj = self.sc.rootFolder
        oSpec.selectSet = self.buildFullTraversal()

        pfs = self.propertyFilterSpec()
        pfs.objectSet = [oSpec]
        pfs.propSet = [
            self.createPropertySpec("VirtualMachine", ["config.uuid", "config.version", "runtime.powerState"]),
            self.createPropertySpec("ClusterComputeResource", ["name"]),
            self.createPropertySpec("HostSystem", ["name",
                                                   "vm",
                                                   "hardware.systemInfo.uuid",
                                                   "hardware.cpuInfo.numCpuPackages",
                                                   "parent",
                                                   "config.product.name",
                                                   "config.product.version",
                                                   "config.network.dnsConfig.hostName",
                                                   "config.network.dnsConfig.domainName"])
        ]

        try:
            return self.client.service.CreateFilter(_this=self.sc.propertyCollector, spec=pfs, partialUpdates=0)
        except requests.RequestException as e:
            raise virt.VirtError(str(e))

    def applyUpdates(self, updateSet):
        for filterSet in updateSet.filterSet:
            for objectSet in filterSet.objectSet:
                if objectSet.obj._type == 'VirtualMachine':  # pylint: disable=W0212
                    self.applyVirtualMachineUpdate(objectSet)
                elif objectSet.obj._type == 'HostSystem':  # pylint: disable=W0212
                    self.applyHostSystemUpdate(objectSet)
                elif objectSet.obj._type == 'ClusterComputeResource':
                    self.applyClusterComputeResource(objectSet)

    def applyClusterComputeResource(self, objectSet):
        if objectSet.kind in ['enter', 'kind']:
            cluster = self.clusters[objectSet.obj.value]
            for change in objectSet.changeSet:
                if change.op == 'assign' and hasattr(change, 'val'):
                    cluster[change.name] = change.val
        elif objectSet.kind == 'leave':
            del self.clusters[objectSet.obj.value]
        else:
            self.logger.error("Unknown update objectSet type: %s", objectSet.kind)

    def applyVirtualMachineUpdate(self, objectSet):
        if objectSet.kind in ['enter', 'modify']:
            vm = self.vms[objectSet.obj.value]
            for change in objectSet.changeSet:
                if change.op == 'assign' and hasattr(change, 'val'):
                    vm[change.name] = change.val
                elif change.op in ['remove', 'indirectRemove']:
                    try:
                        del vm[change.name]
                    except KeyError:
                        pass
                elif change.op == 'add':
                    vm[change.name].append(change.val)
                else:
                    self.logger.error("Unknown change operation: %s", change.op)
        elif objectSet.kind == 'leave':
            del self.vms[objectSet.obj.value]
        else:
            self.logger.error("Unknown update objectSet type: %s", objectSet.kind)

    def applyHostSystemUpdate(self, objectSet):
        if objectSet.kind in ['enter', 'modify']:
            host = self.hosts[objectSet.obj.value]
            for change in objectSet.changeSet:
                if change.op == 'indirectRemove':
                    # Host has been added but without sufficient data
                    # It will be filled in next update
                    pass
                elif change.op == 'assign' and hasattr(change, 'val'):
                    host[change.name] = change.val
        elif objectSet.kind == 'leave':
            del self.hosts[objectSet.obj.value]
        else:
            self.logger.error("Unknown update objectSet type: %s", objectSet.kind)

    def objectSpec(self):
        return self.client.factory.create('ns0:ObjectSpec')

    def traversalSpec(self):
        return self.client.factory.create('ns0:TraversalSpec')

    def selectionSpec(self):
        return self.client.factory.create('ns0:SelectionSpec')

    def propertyFilterSpec(self):
        return self.client.factory.create('ns0:PropertyFilterSpec')

    def buildFullTraversal(self):
        rpToRp = self.createTraversalSpec("rpToRp", "ResourcePool", "resourcePool", ["rpToRp", "rpToVm"])
        rpToVm = self.createTraversalSpec("rpToVm", "ResourcePool", "vm", [])
        crToRp = self.createTraversalSpec("crToRp", "ComputeResource", "resourcePool", ["rpToRp", "rpToVm"])
        crToH = self.createTraversalSpec("crToH", "ComputeResource", "host", [])
        dcToHf = self.createTraversalSpec("dcToHf", "Datacenter", "hostFolder", ["visitFolders"])
        dcToVmf = self.createTraversalSpec("dcToVmf", "Datacenter", "vmFolder", ["visitFolders"])
        hToVm = self.createTraversalSpec("HToVm", "HostSystem", "vm", ["visitFolders"])
        visitFolders = self.createTraversalSpec("visitFolders", "Folder", "childEntity", [
            "visitFolders", "dcToHf", "dcToVmf", "crToH", "crToRp", "HToVm", "rpToVm"])
        return [visitFolders, dcToVmf, dcToHf, crToH, crToRp, rpToRp, hToVm, rpToVm]

    def createPropertySpec(self, type, pathSet, all=False):
        pSpec = self.client.factory.create('ns0:PropertySpec')
        pSpec.all = all
        pSpec.type = type
        pSpec.pathSet = pathSet
        return pSpec

    def createTraversalSpec(self, name, type, path, selectSet):
        ts = self.traversalSpec()
        ts.name = name
        ts.type = type
        ts.path = path
        if len(selectSet) > 0 and isinstance(selectSet[0], str):
            selectSet = self.createSelectionSpec(selectSet)
        ts.selectSet = selectSet
        return ts

    def createSelectionSpec(self, names):
        sss = []
        for name in names:
            ss = self.selectionSpec()
            ss.name = name
            sss.append(ss)
        return sss


class Host(dict):
    def __init__(self):
        self.uuid = None
        self.vms = []


class VM(dict):
    def __init__(self):
        self.uuid = None


class Cluster(dict):
    def __init__(self):
        self.name = None
        self.hosts = []


class EsxConfigSection(VirtConfigSection):
    VIRT_TYPE = 'esx'
    HYPERVISOR_ID = ('uuid', 'hwuuid', 'hostname')

    def __init__(self, *args, **kwargs):
        super(EsxConfigSection, self).__init__(*args, **kwargs)
        self.add_key('server', validation_method=self._validate_server, required=True)
        self.add_key('username', validation_method=self._validate_username, required=True)
        self.add_key('password', validation_method=self._validate_unencrypted_password, required=True)
        self.add_key('simplified_vim', validation_method=self._validate_str_to_bool, default=True)
        self.add_key('filter_host_parents', validation_method=self._validate_filter, default=None)
        self.add_key('exclude_host_parents', validation_method=self._validate_filter, default=None)

    def _validate_server(self, key):
        error = super(EsxConfigSection, self)._validate_server(key)
        if error is None:
            # The suds package makes a specific check to ensure ascii only server names.
            # We must follow that check.
            try:
                self._values[key].encode("ascii")
            except (UnicodeDecodeError, UnicodeEncodeError) as e:
                error = (
                    'error',
                    "Option %s needs to be ASCII characters only: '%s'" % (key, self.name)
                )
        if error is None:
            # Url must contain protocol (usually https://)
            if "://" not in self._values[key]:
                self._values[key] = "https://%s" % self._values[key]
        return error



from __future__ import print_function
"""
Basic module for tests,

Copyright (C) 2017 Radek Novacek <rnovacek@redhat.com>

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""

import logging
from mock import patch, MagicMock
from virtwho.config import VirtConfigSection, ValidationState

# hack to use unittest2 on python <= 2.6, unittest otherwise
# based on python version
import sys
if sys.version_info[0] > 2 or sys.version_info[1] > 6:
    import unittest
else:
    import unittest2 as unittest


class TestBase(unittest.TestCase):
    @classmethod
    @patch('virtwho.log.getLogger')
    def setUpClass(cls, logger):
        cls.logger = logging.getLogger("")
        cls.logger.handlers = []
        cls.logger.setLevel(logging.CRITICAL)
        logger.return_value = cls.logger

    @staticmethod
    def create_fake_config(name, **kwargs):
        # Used to create a fake config with a given name
        # All kwargs that are given will be used as the values provided by the config
        mock_config = MagicMock(spec=VirtConfigSection)
        mock_config.name = name
        mock_config.__getitem__.side_effect = kwargs.__getitem__
        mock_config.__setitem__.side_effect = kwargs.__setitem__
        mock_config.__len__.return_value = kwargs.__len__()
        # Returns the mock_config and the dictionary that underlies it (useful for modifying the
        # contents of the config, without messing up the expected call stack of the mock_config
        return mock_config, kwargs


class ConfigSectionValidationTests(object):
    """
    This is a group of tests which are meant to test that a config section subclass ends up in the
    correct state, after validation, given various inputs.

    Can be used as a mixin to add these tests to any suite (that has valid values for the required
    class attributes).

    This is not a test case itself as we do not expect these tests to pass as written without a
    specific backend CONFIG_CLASS.
    """

    # A dictionary of valid configuration values
    VALID_CONFIG = {}

    # Should be changed for each set of tests, to make sure the right class is utilized
    CONFIG_CLASS = VirtConfigSection

    # should be a list of the keys which are required for sam destination
    SAM_REQUIRED_KEYS = set()

    # should be a list of the keys which are required for sam destination
    SAT5_REQUIRED_KEYS = set()

    # Those keys that should have a default (along with their expected default value)
    DEFAULTS = {}

    def _modified_dict(self, values, excluding=None, including=None):
        """
        Copies a dict removing values where necessary
        Args:
            values: a dict of values to take from
            excluding: a list or set of items to exclude from the copied dict
            including: a list or set of items to include from the copied dict (wins over exclude)

        Returns: a copied and modified dict

        """
        if excluding is None:
            excluding = set()
        if isinstance(excluding, list):
            excluding = set(excluding)

        if including is None:
            including = set()
        if isinstance(including, list):
            including = set(including)

        excluding = excluding - including

        out = {}
        for key in list(values.keys()):
            if key in including or not including:
                out[key] = values[key]
            if key in out and (key in excluding or (including and key not in including)):
                del out[key]
        return out

    def test_valid_config(self):
        """
        This is essentially a smoke test to ensure the following:
        1) An instance of the correct class is created from the valid configuration info
        2) That the configuration is in the proper state pre and post validation
        """
        config = self.CONFIG_CLASS.from_dict(self.VALID_CONFIG, "test", None)

        self.assertIsInstance(config, self.CONFIG_CLASS, 'Wrong type of config, was "type" set'
                                                         'correctly in "VALID_CONFIG"?')

        # Having been updated, the config section needs validation
        self.assertEqual(config.state, ValidationState.NEEDS_VALIDATION)

        messages = config.validate()

        # The config state should be valid, post validation with valid config information
        self.assertEqual(config.state, ValidationState.VALID)

        # There should be no error messages if this config section is valid
        self.assertFalse(any(message[0] == 'error' for message in messages))

    def test_missing_sam_required_no_default(self):
        """
        This tests that when a required key, which has no default, is missing, the config section
        becomes invalid post-validation. This test is specific for SAM destination (candlepin server).
        """
        config = self.CONFIG_CLASS("test", None)
        # A sorted list of keys that are required, but have no default
        target_keys = sorted(list(self.SAM_REQUIRED_KEYS - set(self.DEFAULTS.keys())))
        config_values = self._modified_dict(self.VALID_CONFIG, excluding=target_keys)
        config.update(**config_values)

        messages = config.validate()

        if len(self.SAM_REQUIRED_KEYS) > 0:
            self.assertEqual(config.state, ValidationState.INVALID)
            self.assertTrue(any(message[0] == 'error' for message in messages))

    def test_missing_sat5_required_no_default(self):
        """
        This tests that when a required key, which has no default, is missing, the config section
        becomes invalid post-validation. This test is specific for satellite 5 destination
        """
        config = self.CONFIG_CLASS("test", None)
        # A sorted list of keys that are required, but have no default
        target_keys = sorted(list(self.SAT5_REQUIRED_KEYS - set(self.DEFAULTS.keys())))
        config_values = self._modified_dict(self.VALID_CONFIG, excluding=target_keys)
        config.update(**config_values)

        messages = config.validate()

        if len(self.SAT5_REQUIRED_KEYS) > 0:
            self.assertEqual(config.state, ValidationState.INVALID)
            self.assertTrue(any(message[0] == 'error' for message in messages))

    def test_missing_sam_required_with_default(self):
        """
        Test that a config section given a set of values which are valid other than missing required
        keys which have defaults, will be valid post validation, and that the values for the
        defaults are used.

        NOTE: This does not account for modifications of default values made during validation.
        """
        config = self.CONFIG_CLASS("test", None)
        target_keys = sorted(list(self.SAM_REQUIRED_KEYS & set(self.DEFAULTS.keys())))
        config_values = self._modified_dict(self.VALID_CONFIG, excluding=target_keys)
        config.update(**config_values)

        config.validate()

        self.assertEqual(config.state, ValidationState.VALID)

        for key in target_keys:
            self.assertIn(key, config, "Key '%s' is missing from config")
            self.assertEqual(config[key], self.DEFAULTS[key])



# -*- coding: utf-8 -*-
from __future__ import print_function
"""
For moving variables from system environment to general config file

Copyright (C) 2020 William Poteat <wpoteat@redhat.com>

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""

import os

SYSCONFIG_FILENAME = '/etc/sysconfig/virt-who'
GENERAL_CONFIG_FILENAME = '/etc/virt-who.conf'

def main():
    migrate_env_to_config()

def migrate_env_to_config(sysconfig_filename=None, general_config_filename=None):
    if not sysconfig_filename:
        sysconfig_filename = SYSCONFIG_FILENAME
    if not general_config_filename:
        general_config_filename = GENERAL_CONFIG_FILENAME

    interval = None
    debug = None
    one_shot = None

    # read know env variables
    interval = os.environ.get('VIRTWHO_INTERVAL', None)
    debug = os.environ.get('VIRTWHO_DEBUG', None)
    one_shot = os.environ.get('VIRTWHO_ONE_SHOT', None)

    # read the values in the existing /etc/sysconfig/virt-who file
    # these would override the env variables
    env_vars = {}
    if os.path.exists(sysconfig_filename):
        with open(sysconfig_filename) as myfile:
            for line in myfile:
                if not line.startswith('#'):
                    name, var = line.partition("=")[::2]
                    env_vars[name.strip()] = var

    # pop is used because these values in the global
    # section. The remaining values will go into
    # the system_environment section.
    if 'VIRTWHO_INTERVAL' in env_vars:
        interval = env_vars.pop('VIRTWHO_INTERVAL')
    if 'VIRTWHO_DEBUG' in env_vars:
        debug = env_vars.pop('VIRTWHO_DEBUG')
    if 'VIRTWHO_ONE_SHOT' in env_vars:
        one_shot = env_vars.pop('VIRTWHO_ONE_SHOT')

    # read ini file at /etc/virt-who.conf
    lines = []
    if os.path.exists(general_config_filename):
        with open(general_config_filename) as conf:
            lines = conf.readlines()

    output = []
    has_sys_env = False
    has_global = False

    for line in lines:
        stripped = line.strip()
        if stripped == '[global]' or \
            (stripped.startswith('#') and stripped[1:].strip() == '[global]'):
            has_global = True
            output.append('[global]\n')
            add_global(output, interval, debug, one_shot)
        elif stripped == '[system_environment]' or \
            (stripped.startswith('#') and stripped[1:].strip() == '[system_environment]'):
            has_sys_env = True
            output.append('[system_environment]\n')
            add_system_environment(output, env_vars)
        else:
            output.append(line)

    if not has_global:
        if interval or debug or one_shot:
            if len(output) != 0:
                output.append('\n')
            output.append('[global]\n')
            add_global(output, interval, debug, one_shot)
    if not has_sys_env:
        if len(env_vars) > 0:
            if len(output) != 0:
                output.append('\n')
            output.append('[system_environment]\n')
            add_system_environment(output, env_vars)
    # write /etc/virt-who.conf
    with open(general_config_filename, "w") as conf:
        conf.writelines(output)

def add_global(output, interval, debug, one_shot):
    if interval:
        output.append("#migrated\ninterval=%s\n" % interval.strip())
    if debug:
        output.append("#migrated\ndebug=%s\n" % ('True' if debug.strip()=='1' else 'False'))
    if one_shot:
        output.append("#migrated\noneshot=%s\n" % ('True' if one_shot.strip()=='1' else 'False'))

def add_system_environment(output, env_vars={}):
    for key, value in env_vars.items():
        if key:
            output.append("#migrated\n%s=%s\n" % (key, value.strip()))

if __name__ == '__main__':
    main()


# -*- coding: utf-8 -*-
from __future__ import print_function

import os
import six
import struct
from socket import gethostname
import hmac
import hashlib

if not six.PY3:
    from M2Crypto.RC4 import RC4
else:
    from cryptography.hazmat.primitives.ciphers import Cipher, algorithms
    from cryptography.hazmat.backends import default_backend

    class RC4(object):
        def __init__(self, key):
            if isinstance(key, str):
                key = key.encode('utf-8')
            cipher = Cipher(algorithms.ARC4(key), mode=None, backend=default_backend())
            self.encrypter = cipher.encryptor()

        def update(self, message):
            if isinstance(message, str):
                message = message.encode('utf-8')
            return self.encrypter.update(message)


def rc4k(key, message):
    """
    Compute rc4 of `message` with initial `key`.
    """
    return RC4(key).update(message)


def mac(handle, signing_key, seq_num, message):
    """
    MAC signing method that create signature for given `message` with sequence
    number `seq_num` and using key `signing_key`. The `handle` corresponds to
    current state of sealing key.
    """
    if not isinstance(message, six.binary_type):
        message = message.encode('utf-8')
    hmac_md5 = hmac.new(
        key=signing_key,
        msg=struct.pack('<I', seq_num) + message,
        digestmod=hashlib.md5
    ).digest()[:8]
    checksum = handle.update(hmac_md5)
    return struct.pack('<I8sI', 1, checksum[:8], seq_num)


def nonce(bytes):
    """
    Random data with length of `bytes`.
    """
    return os.urandom(bytes)


def ntlm_compute_response(flags, response_key_nt, response_key_lm,
                          server_challenge, client_challenge, time, target_info):
    """
    Compute NTLMv2 response.

    Return tuple (nt_challenge_response, lm_challenge_response, session_base_key).
    """
    responser_version = b'\x01'
    hi_responser_version = b'\x01'
    temp = (
        responser_version + hi_responser_version + b'\0' * 6 + time +
        client_challenge + b'\0' * 4 + target_info + b'\0' * 4
    )
    nt_proof_str = hmac.new(
        key=response_key_nt,
        msg=server_challenge + temp,
        digestmod=hashlib.md5
    ).digest()
    nt_challenge_response = nt_proof_str + temp
    lm_challenge_response = hmac.new(
        key=response_key_lm,
        msg=server_challenge + client_challenge,
        digestmod=hashlib.md5
    ).digest() + client_challenge
    session_base_key = hmac.new(
        key=response_key_nt,
        msg=nt_proof_str,
        digestmod=hashlib.md5
    ).digest()
    return nt_challenge_response, lm_challenge_response, session_base_key


def ntowfv2(passwd, user, domain):
    """
    Hash password `passwd` using `user` and `domain`.
    """
    return hmac.new(
        key=hashlib.new('md4', passwd.encode('utf-16le')).digest(),
        msg=(user.upper() + domain).encode('utf-16le'),
        digestmod=hashlib.md5
    ).digest()


class NtlmError(Exception):
    pass


# Session keys for NTLM
SESSION_C2S_SEAL = b'session key to client-to-server sealing key magic constant\x00'
SESSION_S2C_SEAL = b'session key to server-to-client sealing key magic constant\x00'
SESSION_C2S_SIGN = b'session key to client-to-server signing key magic constant\x00'
SESSION_S2C_SIGN = b'session key to server-to-client signing key magic constant\x00'

# NTLM Flags
NTLM_NegotiateUnicode = 0x00000001
NTLM_NegotiateOEM = 0x00000002
NTLM_RequestTarget = 0x00000004
NTLM_Unknown9 = 0x00000008
NTLM_NegotiateSign = 0x00000010
NTLM_NegotiateSeal = 0x00000020
NTLM_NegotiateDatagram = 0x00000040
NTLM_NegotiateLanManagerKey = 0x00000080
NTLM_Unknown8 = 0x00000100
NTLM_NegotiateNTLM = 0x00000200
NTLM_NegotiateNTOnly = 0x00000400
NTLM_Anonymous = 0x00000800
NTLM_NegotiateOemDomainSupplied = 0x00001000
NTLM_NegotiateOemWorkstationSupplied = 0x00002000
NTLM_Unknown6 = 0x00004000
NTLM_NegotiateAlwaysSign = 0x00008000
NTLM_TargetTypeDomain = 0x00010000
NTLM_TargetTypeServer = 0x00020000
NTLM_TargetTypeShare = 0x00040000
NTLM_NegotiateExtendedSecurity = 0x00080000
NTLM_NegotiateIdentify = 0x00100000
NTLM_Unknown5 = 0x00200000
NTLM_RequestNonNTSessionKey = 0x00400000
NTLM_NegotiateTargetInfo = 0x00800000
NTLM_Unknown4 = 0x01000000
NTLM_NegotiateVersion = 0x02000000
NTLM_Unknown3 = 0x04000000
NTLM_Unknown2 = 0x08000000
NTLM_Unknown1 = 0x10000000
NTLM_Negotiate128 = 0x20000000
NTLM_NegotiateKeyExchange = 0x40000000
NTLM_Negotiate56 = 0x80000000


class Message(object):
    """
    Abstract message for NTLM authentication.
    """
    VERSION = [
        ('product_major_version', '<B'),
        ('product_minor_version', '<B'),
        ('product_build', '<H'),
        ('reserved', '<B'),
        ('reserved', '<B'),
        ('reserved', '<B'),
        ('ntlm_revision_current', '<B'),
    ]

    DEFAULTS = {
        'signature': b'NTLMSSP\0',
        'product_major_version': 5,
        'product_minor_version': 1,
        'product_build': 2600,
        'reserved': 0,
        'ntlm_revision_current': 15,
    }


class IncomingMessage(Message):
    """
    Abstract incoming NTLM message.
    """
    def __init__(self, message):
        self._items = {}
        self.payload = None
        self._parse(message)

    def _parse(self, message):
        offset = 0
        for name, signature in self.FORMAT:
            self._items[name] = struct.unpack_from(signature, message, offset)[0]
            offset += struct.calcsize(signature)
        self.payload = message[offset:]

    def __getattr__(self, name):
        try:
            return self._items[name]
        except KeyError:
            return Message.__getattr__(self, name)


class OutgoingMessage(Message):
    """
    Abstract outgoing NTLM message.
    """
    def __init__(self, params):
        self.params = params

    def _format(self):
        message = []
        for name, signature in self.FORMAT:
            item = self.params.get(name)
            if item is None:
                item = self.DEFAULTS[name]
            if 's' in signature and isinstance(item, str):
                item = item.encode('utf-8')
            message.append(struct.pack(signature, item))
        res = message[0]
        for item in message[1:]:
            res += item
        return res

    @property
    def data(self):
        return self._format()


class NegotiateMessage(OutgoingMessage):
    """
    NegotiateMessage is first message (sent by client) to initiate NTLM auth.
    """
    HEADER_LENGTH = 40
    FORMAT = [
        ('signature', '8s'),
        ('message_type', '<I'),
        ('negotiate_flags', '<I'),
        ('domain_name_len', '<H'),
        ('domain_name_len', '<H'),
        ('domain_name_buffer_offset', '<I'),
        ('workstation_len', '<H'),
        ('workstation_len', '<H'),
        ('workstation_buffer_offset', '<I'),
    ] + OutgoingMessage.VERSION
    DEFAULTS = OutgoingMessage.DEFAULTS.copy()
    DEFAULTS.update({
        'message_type': 0x00000001,
        'negotiate_flags': (
            NTLM_NegotiateUnicode |
            NTLM_NegotiateOEM |
            NTLM_RequestTarget |
            NTLM_NegotiateNTLM |
            NTLM_NegotiateOemWorkstationSupplied |
            NTLM_NegotiateAlwaysSign |
            NTLM_NegotiateSign |
            NTLM_NegotiateSeal |
            NTLM_NegotiateExtendedSecurity |
            NTLM_NegotiateVersion |
            NTLM_Negotiate128 |
            NTLM_Negotiate56 |
            NTLM_NegotiateKeyExchange
        ),
    })

    def __init__(self, domain, workstation, flags=None):
        self.domain = domain if not six.PY3 else domain.encode('ascii')
        self.workstation = workstation if not six.PY3 else workstation.encode('ascii')
        self.flags = flags

    def _format(self):
        domain_len = len(self.domain)
        domain_offset = self.HEADER_LENGTH
        workstation_len = len(self.workstation)
        workstation_offset = domain_offset + domain_len
        self.params = {
            'domain_name_len': domain_len,
            'domain_name_buffer_offset': domain_offset,
            'workstation_len': workstation_len,
            'workstation_buffer_offset': workstation_offset
        }
        if self.flags is not None:
            self.params['negotiate_flags'] = self.flags
        else:
            self.params['negotiate_flags'] = self.DEFAULTS['negotiate_flags']
        if domain_len > 0:
            self.params['negotiate_flags'] |= NTLM_NegotiateOemDomainSupplied
        if six.PY3:
            return OutgoingMessage._format(self) + self.domain + self.workstation
        else:
            return OutgoingMessage._format(self) + self.domain.encode('utf-8') + self.workstation.encode('utf-8')


class ChallengeMessage(IncomingMessage):
    """
    ChallengeMessage is send by server to provide data for NTLM authentication
    and sealing.
    """
    FORMAT = (
        ('signature', '8s'),
        ('message_type', '<I'),
        ('target_name_len', '<H'),
        ('target_name_len_max', '<H'),
        ('target_name_buffer_offset', '<I'),
        ('negotiate_flags', '<I'),
        ('server_challenge', '8s'),
        ('reserved', '8s'),
        ('target_info_len', '<H'),
        ('target_info_len_max', '<H'),
        ('target_info_buffer_offset', '<I'),
        ('version', '8s'),
        # Payload (variable)
    )

    def _parse(self, message):
        IncomingMessage._parse(self, message)
        assert self._items['signature'] == b'NTLMSSP\x00'
        assert self._items['message_type'] == 2
        self._items['target_name'] = message[
            self._items['target_name_buffer_offset']:
            self._items['target_name_buffer_offset'] + self._items['target_name_len']]
        self._items['target_info'] = message[
            self._items['target_info_buffer_offset']:
            self._items['target_info_buffer_offset'] + self._items['target_info_len']]
        flags = self._items['negotiate_flags']
        for flag in (NTLM_NegotiateUnicode, NTLM_NegotiateExtendedSecurity, NTLM_Negotiate128):
            if not flag & flags:
                raise NtlmError("NTLM negotiation failed, no flag %d" % flag)


class AuthenticationMessage(OutgoingMessage):
    """
    AuthenticateMessage is final message that the client sends. It contains
    encrypted session key and other data used by authentication and sealing.
    """
    HEADER_LENGTH = 72
    FORMAT = [
        ('signature', '8s'),
        ('message_type', '<I'),
        ('lm_challenge_response_len', '<H'),
        ('lm_challenge_response_len', '<H'),
        ('lm_challenge_response_buffer_offset', '<I'),
        ('nt_challenge_response_len', '<H'),
        ('nt_challenge_response_len', '<H'),
        ('nt_challenge_response_buffer_offset', '<I'),
        ('domain_name_len', '<H'),
        ('domain_name_len', '<H'),
        ('domain_name_buffer_offset', '<I'),
        ('user_name_len', '<H'),
        ('user_name_len', '<H'),
        ('user_name_buffer_offset', '<I'),
        ('workstation_len', '<H'),
        ('workstation_len', '<H'),
        ('workstation_buffer_offset', '<I'),
        ('encrypted_random_session_key_len', '<H'),
        ('encrypted_random_session_key_len', '<H'),
        ('encrypted_random_session_key_buffer_offset', '<I'),
        ('negotiate_flags', '<I'),
    ] + OutgoingMessage.VERSION
    DEFAULTS = OutgoingMessage.DEFAULTS.copy()
    DEFAULTS.update({
        'message_type': 0x00000003,
        'negotiate_flags': (
            NTLM_NegotiateKeyExchange |
            NTLM_Negotiate128 |
            NTLM_Negotiate56 |
            NTLM_NegotiateVersion |
            NTLM_NegotiateTargetInfo |
            NTLM_NegotiateExtendedSecurity |
            NTLM_NegotiateUnicode |
            NTLM_NegotiateSign |
            NTLM_NegotiateSeal |
            NTLM_NegotiateNTLM |
            NTLM_NegotiateAlwaysSign |
            NTLM_RequestTarget
        ),
    })

    def __init__(self, username, password, domain, workstation, server_challenge,
                 target_info, negotiate_flags, client_challenge=None,
                 exported_session_key=None):
        self.username = username
        self.password = password
        self.domain = domain
        self.workstation = workstation
        self.server_challenge = server_challenge
        self.target_info = target_info
        self.time = self._time_from_target_info(target_info)
        self.negotiate_flags = negotiate_flags
        if client_challenge is not None:
            self.client_challenge = client_challenge
        else:
            self.client_challenge = nonce(8)
        if exported_session_key is not None:
            self.exported_session_key = exported_session_key
        else:
            self.exported_session_key = nonce(16)

        self._compute_encryption_data()

    def _compute_encryption_data(self):
        """
        Compute data that are needed for authentication and encryption.
        """
        time = self.time or b'\0' * 8

        response_key_nt = response_key_lm = ntowfv2(self.password, self.username, self.domain)
        self.nt_challenge_response, self.lm_challenge_response, session_base_key = ntlm_compute_response(
            self.negotiate_flags, response_key_nt, response_key_lm,
            self.server_challenge, self.client_challenge, time, self.target_info)

        if self.time:
            # Send NULLs instead of lm_challenge_response if we have time from server
            self.lm_challenge_response = b'\0' * 24

        key_exchange_key = session_base_key  # key_exchange_key is always session_base_key in NTLMv2
        if self.negotiate_flags & NTLM_NegotiateKeyExchange:
            exported_session_key = self.exported_session_key
            self.encrypted_random_session_key = rc4k(key_exchange_key, exported_session_key)
        else:
            exported_session_key = key_exchange_key
            self.encrypted_random_session_key = ''

        # The NTLM_NegotiateExtendedSecurity flag is always set
        if NTLM_Negotiate128 & self.negotiate_flags:
            self.session_key = exported_session_key
        elif NTLM_Negotiate56 & self.negotiate_flags:
            self.session_key = exported_session_key[:7]
        else:
            self.session_key = exported_session_key[:4]

    def _time_from_target_info(self, target_info):
        """
        Extract timestamp from target_info.
        """
        l = len(target_info)
        offset = 0
        timestamp = None
        while offset < l:
            av_id = struct.unpack_from('<H', target_info, offset)[0]
            offset += 2
            av_len = struct.unpack_from('<H', target_info, offset)[0]
            offset += 2
            if av_id == 0x0007:  # Timestamp
                timestamp = target_info[offset:offset + av_len]
            offset += av_len
        return timestamp

    def _format(self):
        username = self.username.encode('utf-16-le')
        domain = self.domain.encode('utf-16-le')
        workstation = self.workstation.encode('utf-16-le')
        domain_len = len(domain)
        domain_offset = self.HEADER_LENGTH
        username_len = len(username)
        username_offset = domain_offset + domain_len
        workstation_len = len(workstation)
        workstation_offset = username_offset + username_len

        lm_challenge_response_len = len(self.lm_challenge_response)
        lm_challenge_response_offset = workstation_offset + workstation_len

        nt_challenge_response_len = len(self.nt_challenge_response)
        nt_challenge_response_offset = lm_challenge_response_offset + lm_challenge_response_len

        encrypted_random_session_key_len = len(self.encrypted_random_session_key)
        encrypted_random_session_key_offset = nt_challenge_response_offset + nt_challenge_response_len
        mic = b'\0' * 16

        flags = self.DEFAULTS['negotiate_flags']
        self.params = {
            'lm_challenge_response_len': lm_challenge_response_len,
            'lm_challenge_response_buffer_offset': lm_challenge_response_offset,
            'nt_challenge_response_len': nt_challenge_response_len,
            'nt_challenge_response_buffer_offset': nt_challenge_response_offset,
            'domain_name_len': domain_len,
            'domain_name_buffer_offset': domain_offset,
            'user_name_len': username_len,
            'user_name_buffer_offset': username_offset,
            'workstation_len': workstation_len,
            'workstation_buffer_offset': workstation_offset,
            'encrypted_random_session_key_len': encrypted_random_session_key_len,
            'encrypted_random_session_key_buffer_offset': encrypted_random_session_key_offset,
            'negotiate_flags': flags,
            'mic': mic,
        }
        return (
            OutgoingMessage._format(self) + domain + username + workstation +
            self.lm_challenge_response + self.nt_challenge_response +
            self.encrypted_random_session_key)


class Ntlm(object):
    """
    Wrapper for NTLM authentication and sealing.

    Usage:
    First call `negotiate_message` to create message of type 1 to be send
    to server. Then server will respond with message of type 2. Supply it to
    `authentication_message` as `challenge` argument together with `password`.
    This method will return type 3 message that should be send to server
    together with (possibly encrypted) data.

    After these steps, you can encrypt and decrypt message using `encrypt`
    and `decrypt` methods.
    """
    def __init__(self):
        self.incoming_seq_number = 0
        self.outgoing_seq_number = 0

    def negotiate_message(self, username):
        """
        Create type 1 message to be send to server.
        """
        user_parts = username.split('\\', 1)
        if len(user_parts) > 1:
            self.domain = user_parts[0].upper()
            self.username = user_parts[1]
        else:
            self.domain = ''
            self.username = username
        self.workstation = gethostname().upper()
        data = NegotiateMessage(self.domain, self.workstation).data
        return data

    def authentication_message(self, challenge, password):
        """
        Create type 3 message from type 2 (`challenge` argument) and user
        `password`.
        """
        challenge = ChallengeMessage(challenge)
        msg = AuthenticationMessage(
            self.username, password, self.domain, self.workstation,
            challenge.server_challenge, challenge.target_info,
            challenge.negotiate_flags)

        self.set_session_key(msg.session_key)
        return msg.data

    def set_session_key(self, session_key):
        """
        Set session key that will be used for encryption and decryption.

        If you call `authentication_message` you don't need to call this method.
        """
        self.session_key = session_key

        self.outgoing_sealing_key = hashlib.md5(session_key + SESSION_C2S_SEAL).digest()
        self.incoming_sealing_key = hashlib.md5(session_key + SESSION_S2C_SEAL).digest()
        self.outgoing_signing_key = hashlib.md5(session_key + SESSION_C2S_SIGN).digest()
        self.incoming_signing_key = hashlib.md5(session_key + SESSION_S2C_SIGN).digest()

        self.outgoing_seal_handle = RC4(self.outgoing_sealing_key)
        self.incoming_seal_handle = RC4(self.incoming_sealing_key)

    def encrypt(self, message):
        """
        Encrypt and sign given `message` and return pair
        (encrypted_message, signature).
        """
        sealed_message = self.outgoing_seal_handle.update(message)
        signature = mac(self.outgoing_seal_handle, self.outgoing_signing_key, self.outgoing_seq_number, message)
        self.outgoing_seq_number += 1
        return sealed_message, signature

    def decrypt(self, sealed_message, signature):
        """
        Decrypt `sealed_message` and check it signature. Return decrypted
        message or Exception if sequence number or signature doesn't match.
        """
        message = self.incoming_seal_handle.update(sealed_message)
        version, checksum, sequence = struct.unpack('<I8sI', signature)
        if sequence != self.incoming_seq_number:
            raise Exception("Incorrect sequence number")
        checksum = self.incoming_seal_handle.update(checksum)
        expected_checksum = hmac.new(
            key=self.incoming_signing_key,
            msg=struct.pack('<I', self.incoming_seq_number) + message,
            digestmod=hashlib.md5
        ).digest()[:8]
        self.incoming_seq_number += 1
        if checksum != expected_checksum:
            raise Exception("Message has been altered")
        return message



from __future__ import print_function

#
# Module for abstraction of all virtualization backends, part of virt-who
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

import sys
import time

from fake_virt import FakeVirt
from fake_server import FakeHandler


class Rhevm4Handler(FakeHandler):
    def do_GET(self):
        time.sleep(0.1)
        print(("[RHEVM] DO GET", self.path))

        if self.path == '/ovirt-engine/api/v4':
            self.write_file('rhevm', 'rhev4_api.xml')
        elif self.path == '/ovirt-engine/api/v4/clusters':
            self.write_file('rhevm', 'rhevm_clusters.xml')
        elif self.path == '/ovirt-engine/api/v4/hosts':
            self.write_file('rhevm', 'rhevm_hosts.xml')
        elif self.path == '/ovirt-engine/api/v4/vms':
            self.write_file('rhevm', 'rhevm_vms_%d.xml' % self.server._data_version.value)
        else:
            self.send_response(404)
            self.end_headers()


class FakeRhevm4(FakeVirt):

    virt_type = "rhevm"

    def __init__(self, port=None):
        super(FakeRhevm4, self).__init__(Rhevm4Handler, port=port)
        self.server._data_version = self._data_version


if __name__ == '__main__':
    if len(sys.argv) >= 2:
        port = int(sys.argv[1])
    else:
        port = None
    rhevm = FakeRhevm4(port=port)
    rhevm.run()



from __future__ import print_function

#
# Module for abstraction of all virtualization backends, part of virt-who
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

import os
import tempfile
import shutil

import virtwhotest

from fake_esx import FakeEsx


class EsxTest(virtwhotest.TestBase):
    """
    Class for complex testing of obtaining host-to-guest mapping from fake ESX server
    """

    virt = 'esx'

    def __init__(self, *args, **kwargs):
        super(EsxTest, self).__init__(*args, **kwargs)

    @classmethod
    def setUpClass(cls):
        super(EsxTest, cls).setUpClass()
        cls.server = FakeEsx()
        cls.server.start()
        cls.config_dir = tempfile.mkdtemp()
        with open(os.path.join(cls.config_dir, "test.conf"), "w") as f:
            config_file_content = """
[test]
type=esx
server=http://localhost:%s
username=%s
password=%s
owner=admin

rhsm_hostname=localhost
rhsm_port=%s
rhsm_username=admin
rhsm_password=admin
rhsm_org=admin
rhsm_prefix=/
rhsm_insecure=1
""" % (cls.server.port, cls.server.username, cls.server.password, cls.sam.port)
            print("Using config: %s" % config_file_content)
            f.write(config_file_content)

        cls.arguments = [
            '-c=%s' % os.path.join(cls.config_dir, "test.conf")
        ]

    @classmethod
    def tearDownClass(cls):
        super(EsxTest, cls).tearDownClass()
        cls.server.terminate()
        cls.server.join()
        shutil.rmtree(cls.config_dir)


SERVER_BASE_URIL = 'https://%s:%d/api/nutanix/%s'
AHV_HYPERVIRSOR = ['kKvm', 'AHV', 'ahv', 'kvm']
TASK_COMPLETE_MSG = ['SUCCEEDED', 'Succeeded']
DEFAULT_PORT = 9440
VERSION_2 = 'v2.0'
VERSION_3 = 'v3'

CMN_RST_CMD = {'get_vm': {'url': '/vms/%s', 'method': 'get'},
               'get_host': {'url': '/hosts/%s', 'method': 'get'},
               'get_tasks': {'url': '/tasks/list', 'method': 'post'},
               'get_task': {'url': '/tasks/%s', 'method': 'get'}}

REST_CMD = {VERSION_2: {'list_vms': {'url': '/vms', 'method': 'get'},
                        'list_hosts': {'url': '/hosts', 'method': 'get'},
                        'list_clusters' : {'url': '/clusters',
                                           'method': 'get'}},
            VERSION_3: {'list_vms': {'url': '/vms/list', 'method': 'post'},
                        'list_hosts': {'url': '/hosts/list', 'method': 'post'},
                        'list_clusters' : {'url': '/clusters/list',
                                           'method': 'post'}}}




from __future__ import print_function, absolute_import
"""
Test for Satellite module, part of virt-who

Copyright (C) 2013 Radek Novacek <rnovacek@redhat.com>

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""
import os
import sys

import threading
import tempfile
import pickle
import shutil
import six
import pytest

from six.moves import xmlrpc_client
from six.moves.xmlrpc_server import SimpleXMLRPCServer, SimpleXMLRPCRequestHandler
from binascii import hexlify
from mock import Mock, MagicMock, patch

from base import TestBase

from virtwho.config import DestinationToSourceMapper, EffectiveConfig, ConfigSection,\
    parse_file, Satellite5DestinationInfo, VirtConfigSection
from virtwho.manager import Manager
from virtwho.manager.satellite import Satellite, SatelliteError
from virtwho.virt import Guest, Hypervisor, HostGuestAssociationReport
from virtwho.parser import parse_options
from virtwho import password


TEST_SYSTEM_ID = 'test-system-id'
TEST_PORT = 8090


class RequestHandler(SimpleXMLRPCRequestHandler):
    rpc_paths = ('/XMLRPC', '/rpc/api')


class FakeSatellite(SimpleXMLRPCServer):
    AUTH_TOKEN = 'This is auth session'

    def __init__(self):
        SimpleXMLRPCServer.__init__(self, ("localhost", TEST_PORT), requestHandler=RequestHandler)
        # /XMLRPC interface
        self.register_function(self.new_system_user_pass, "registration.new_system_user_pass")
        self.register_function(self.refresh_hw_profile, "registration.refresh_hw_profile")
        self.register_function(self.virt_notify, "registration.virt_notify")
        # /xml/api interface
        self.register_function(self.auth_login, "auth.login")
        self.register_function(self.get_channel_details, "channel.software.getDetails")
        self.register_function(self.create_channel, "channel.software.create")
        self.register_function(self.set_map_for_org, "distchannel.setMapForOrg")
        self.register_function(self.get_user_details, "user.getDetails")

        self.channel_created = False
        self.created_system = None

    def new_system_user_pass(self, profile_name, os_release_name, version, arch, username, password, options):
        if username != "username":
            raise Exception("Wrong username")
        if password != "password":
            raise Exception("Wrong password")
        self.created_system = {
            'profile_name': profile_name,
            'os_release_name': os_release_name,
            'version': version,
            'arch': arch,
            'username': username,
            'password': password,
            'options': options,
        }
        return {'system_id': TEST_SYSTEM_ID}

    def refresh_hw_profile(self, system_id, profile):
        if system_id != TEST_SYSTEM_ID:
            raise Exception("Wrong system id")
        return ""

    def virt_notify(self, system_id, plan):
        if system_id != TEST_SYSTEM_ID:
            raise xmlrpc_client.Fault(-9, "Wrong system id")

        if plan[0] != [0, 'exists', 'system', {'uuid': '0000000000000000', 'identity': 'host'}]:
            raise Exception("Wrong value for virt_notify: invalid format of first entry")
        if plan[1] != [0, 'crawl_began', 'system', {}]:
            raise Exception("Wrong value for virt_notify: invalid format of second entry")
        if plan[-1] != [0, 'crawl_ended', 'system', {}]:
            raise Exception("Wrong value for virt_notify: invalid format of last entry")
        for item in plan[2:-1]:
            if item[0] != 0:
                raise Exception("Wrong value for virt_notify: invalid format first item of an entry")
            if item[1] != 'exists':
                raise Exception("Wrong value for virt_notify: invalid format second item of an entry")
            if item[2] != 'domain':
                raise Exception("Wrong value for virt_notify: invalid format third item of an entry")
            if not item[3]['uuid'].startswith("guest"):
                raise Exception("Wrong value for virt_notify: invalid format uuid item")
        return 0

    def auth_login(self, username, password):
        return self.AUTH_TOKEN

    def get_channel_details(self, session, channelLabel):
        assert session == self.AUTH_TOKEN
        if self.channel_created:
            return {
                'id': 42
            }
        else:
            raise xmlrpc_client.Fault(faultCode=-210, faultString='Not found')

    def create_channel(self, session, label, name, summary, archLabel, parentLabel):
        assert session == self.AUTH_TOKEN
        self.channel_created = True
        return 1

    def set_map_for_org(self, session, os, release, archName, channelLabel):
        assert session == self.AUTH_TOKEN
        return 1

    def get_user_details(self, session, login):
        assert session == self.AUTH_TOKEN
        return dict(org_id=101)

xvirt = type("", (), {'CONFIG_TYPE': 'xxx'})()


class TestSatellite(TestBase):
    mapping = {
        'hypervisors': [
            Hypervisor('host-1', [
                Guest('guest1-1', xvirt.CONFIG_TYPE, Guest.STATE_RUNNING),
                Guest('guest1-2', xvirt.CONFIG_TYPE, Guest.STATE_SHUTOFF)
            ]),
            Hypervisor('host-2', [
                Guest('guest2-1', xvirt.CONFIG_TYPE, Guest.STATE_RUNNING),
                Guest('guest2-2', xvirt.CONFIG_TYPE, Guest.STATE_SHUTOFF),
                Guest('guest2-3', xvirt.CONFIG_TYPE, Guest.STATE_RUNNING)
            ])
        ]
    }

    # Used as the default configuration
    # Override in other tests if need be
    default_config_args = {
        'type': 'libvirt',
        'hypervisor_id': 'uuid',
        'simplified_vim': True,
        'sat_server': "http://localhost:%s" % TEST_PORT,
        'sat_username': 'username',
        'sat_password': 'password',
    }

    @classmethod
    def setUpClass(cls):
        super(TestSatellite, cls).setUpClass()
        cls.fake_server = FakeSatellite()
        cls.thread = threading.Thread(target=cls.fake_server.serve_forever)
        cls.thread.start()

    @classmethod
    def tearDownClass(cls):
        cls.fake_server.shutdown()

    def test_wrong_server(self):
        options = Mock()
        s = Satellite(self.logger, options)
        config, d = self.create_fake_config('test', **TestSatellite.default_config_args)
        d['sat_server'] = "wrong_server"
        d['sat_username'] = "abc"
        d['sat_password'] = "def"
        report = HostGuestAssociationReport(config, self.mapping)
        self.assertRaises(SatelliteError, s.hypervisorCheckIn, report, options)

    def test_wrong_username(self):
        options = Mock()
        options.force_register = True
        s = Satellite(self.logger, options)
        config, d = self.create_fake_config('test', **TestSatellite.default_config_args)
        d['sat_server'] = "http://localhost:%s" % TEST_PORT
        d['sat_username'] = "wrong"
        d['sat_password'] = "password"
        report = HostGuestAssociationReport(config, self.mapping)
        self.assertRaises(SatelliteError, s.hypervisorCheckIn, report, options)

    def test_wrong_password(self):
        options = Mock()
        options.force_register = True
        s = Satellite(self.logger, options)
        config, d = self.create_fake_config('test', **TestSatellite.default_config_args)
        d['sat_server'] = "http://localhost:%s" % TEST_PORT
        d['sat_username'] = "username"
        d['sat_password'] = "wrong"
        report = HostGuestAssociationReport(config, self.mapping)
        self.assertRaises(SatelliteError, s.hypervisorCheckIn, report, options)

    def test_new_system(self):
        options = Mock()
        options.force_register = True
        s = Satellite(self.logger, options)

        config, d = self.create_fake_config('test', **TestSatellite.default_config_args)
        report = HostGuestAssociationReport(config, self.mapping)
        s.hypervisorCheckIn(report, options)

    def test_hypervisor_checkin(self):
        options = Mock()
        options.force_register = True
        s = Satellite(self.logger, options)

        config, d = self.create_fake_config('test', **TestSatellite.default_config_args)
        report = HostGuestAssociationReport(config, self.mapping)
        result = s.hypervisorCheckIn(report, options)
        self.assertTrue("failedUpdate" in result)
        self.assertTrue("created" in result)
        self.assertTrue("updated" in result)

    def test_hypervisorCheckIn_preregistered(self):
        temp, filename = tempfile.mkstemp(suffix=TEST_SYSTEM_ID)
        self.addCleanup(os.unlink, filename)
        f = os.fdopen(temp, "wb")
        pickle.dump({'system_id': TEST_SYSTEM_ID}, f)
        f.close()

        options = Mock()
        s = Satellite(self.logger, options)

        s.HYPERVISOR_SYSTEMID_FILE = filename.replace(TEST_SYSTEM_ID, '%s')

        config, d = self.create_fake_config('test', **TestSatellite.default_config_args)
        report = HostGuestAssociationReport(config, self.mapping)
        result = s.hypervisorCheckIn(report, options)
        self.assertTrue("failedUpdate" in result)
        self.assertTrue("created" in result)
        self.assertTrue("updated" in result)

    def test_hypervisorCheckIn_deleted(self):
        '''Test running hypervisorCheckIn on system that was deleted from Satellite'''
        system_id = 'wrong-system-id'
        temp, filename = tempfile.mkstemp(suffix=system_id)
        self.addCleanup(os.unlink, filename)
        with os.fdopen(temp, "wb") as f:
            pickle.dump({'system_id': system_id}, f)

        options = Mock()
        s = Satellite(self.logger, options)

        s.HYPERVISOR_SYSTEMID_FILE = filename.replace(system_id, '%s')
        config, d = self.create_fake_config('test', **TestSatellite.default_config_args)
        mapping = {
            'hypervisors': [
                Hypervisor(system_id, [])
            ]
        }
        report = HostGuestAssociationReport(config, mapping)
        s.hypervisorCheckIn(report, options)
        with open(filename, "rb") as f:
            data = pickle.load(f)
        self.assertEqual(data['system_id'], TEST_SYSTEM_ID)

    def test_creating_channel(self):
        # TODO Remove Options entirely
        options = Mock()
        options.force_register = True
        s = Satellite(self.logger, options)

        config, d = self.create_fake_config('test', **TestSatellite.default_config_args)
        report = HostGuestAssociationReport(config, self.mapping)
        result = s.hypervisorCheckIn(report, options)
        self.assertTrue(self.fake_server.channel_created)
        self.assertIsNotNone(self.fake_server.created_system)
        self.assertTrue("created" in result)

    def test_using_existing_channel(self):
        options = Mock()
        options.force_register = True
        s = Satellite(self.logger, options)
        self.fake_server.channel_created = True

        config, d = self.create_fake_config('test', **TestSatellite.default_config_args)
        report = HostGuestAssociationReport(config, self.mapping)
        result = s.hypervisorCheckIn(report, options)
        self.assertTrue(self.fake_server.channel_created)
        self.assertIsNotNone(self.fake_server.created_system)
        self.assertTrue("created" in result)

    def test_per_config_options(self):
        options = Mock()
        options.force_register = True
        config, d = self.create_fake_config('test', **TestSatellite.default_config_args)
        s = Satellite(self.logger, options)

        report = HostGuestAssociationReport(config, self.mapping)
        result = s.hypervisorCheckIn(report, options)
        self.assertTrue("failedUpdate" in result)
        self.assertTrue("created" in result)
        self.assertTrue("updated" in result)

    @patch('virtwho.password.Password._can_write')
    def test_per_config_options_encrypted(self, can_write):
        options = Mock()
        options.force_register = True
        can_write.return_value = True
        with tempfile.NamedTemporaryFile() as tmp:
            password.Password.KEYFILE = tmp.name
            config_dict = {
                "sat_server": "http://localhost:%s" % TEST_PORT,
                "sat_username": "username",
                "sat_encrypted_password": hexlify(password.Password.encrypt('password')),
                "type": "libvirt",
            }
            config = VirtConfigSection.from_dict(config_dict, 'test', None)
            config.validate()
            dests = DestinationToSourceMapper.parse_dests_from_dict(config._values)
            self.assertEqual(len(dests), 1)
            dest_info = dests.pop()
            s = Manager.fromInfo(self.logger, options, dest_info)
            self.assertTrue(isinstance(s, Satellite))
            report = HostGuestAssociationReport(config, self.mapping)
            result = s.hypervisorCheckIn(report, options)
        self.assertTrue("failedUpdate" in result)
        self.assertTrue("created" in result)
        self.assertTrue("updated" in result)


class TestSatelliteConfig(TestBase):

    def setUp(self):
        self.config_dir = tempfile.mkdtemp()
        self.addCleanup(shutil.rmtree, self.config_dir)
        conf_dir_patch = patch('virtwho.config.VW_CONF_DIR', self.config_dir)
        conf_dir_patch.start()
        self.addCleanup(conf_dir_patch.stop)

    @pytest.mark.skipif(not six.PY2, reason="test only runs with python 2 virt-who")
    def test_satellite_config_env(self):
        os.environ = {
            "VIRTWHO_SATELLITE": '1',
            "VIRTWHO_SATELLITE_SERVER": 'sat.example.com',
            "VIRTWHO_SATELLITE_USERNAME": 'username',
            "VIRTWHO_SATELLITE_PASSWORD": 'password',
            "VIRTWHO_LIBVIRT": '1'
        }
        sys.argv = ["virt-who"]
        logger, effective_config = parse_options()
        config_manager = DestinationToSourceMapper(effective_config)
        # Again there should only be one config parsed out (and one dest)
        self.assertEqual(len(config_manager.configs), 1)
        self.assertEqual(len(config_manager.dests), 1)
        dest_info = config_manager.dests.pop()
        self.assertTrue(isinstance(dest_info, Satellite5DestinationInfo))
        manager = Manager.fromInfo(self.logger, effective_config, dest_info)
        self.assertTrue(isinstance(manager, Satellite))

    @pytest.mark.skipif(not six.PY2, reason="test only runs with python 2 virt-who")
    def test_satellite_config_cmd(self):
        os.environ = {}
        sys.argv = ["virt-who", "--satellite",
                    "--satellite-server=sat.example.com",
                    "--satellite-username=username",
                    "--satellite-password=password",
                    "--libvirt"]
        logger, effective_config = parse_options()
        config_manager = DestinationToSourceMapper(effective_config)
        # Again there should only be one config parsed out (and one dest)
        self.assertEqual(len(config_manager.configs), 1)
        self.assertEqual(len(config_manager.dests), 1)
        dest_info = config_manager.dests.pop()
        self.assertTrue(isinstance(dest_info, Satellite5DestinationInfo))
        manager = Manager.fromInfo(self.logger, effective_config, dest_info)
        self.assertTrue(isinstance(manager, Satellite))

    def test_satellite_config_file(self):
        # Username and password are required for a valid sat5 destination
        with open(os.path.join(self.config_dir, "test.conf"), "w") as f:
            f.write("""
[test]
type=libvirt
sat_server=sat.example.com
sat_username=sat_username
sat_password=sat_password
""")
        conf = parse_file(os.path.join(self.config_dir, "test.conf"))
        effective_config = EffectiveConfig()
        conf_values = conf.pop("test")
        effective_config["test"] = ConfigSection.from_dict(
            conf_values,
            "test",
            effective_config
        )
        config_manager = DestinationToSourceMapper(effective_config)
        self.assertEqual(len(config_manager.configs), 1)
        # There should only be one destination detected
        self.assertEqual(len(config_manager.dests), 1)
        # Which should be a Satellite5DestinationInfo
        dest_info = config_manager.dests.pop()
        self.assertTrue(isinstance(dest_info, Satellite5DestinationInfo), 'The destination info '
                                                                          'we got was not of the '
                                                                          'expected type')
        manager = Manager.fromInfo(self.logger, effective_config, dest_info)
        self.assertTrue(isinstance(manager, Satellite))
        self.assertEqual(dest_info.sat_server, 'sat.example.com')



# -*- coding: utf-8 -*-
from __future__ import print_function
"""
Module for communication with libvirt, part of virt-who

Copyright (C) 2014 Radek Novacek <rnovacek@redhat.com>

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""

import time
import libvirt
import threading
from six.moves import urllib
from xml.etree import ElementTree

from virtwho.virt import (
    Hypervisor, Guest, VirtError, HostGuestAssociationReport,
    DomainListReport, Virt, StatusReport)
from virtwho.config import VirtConfigSection


class LibvirtdConfigSection(VirtConfigSection):
    """
    This class is used for validation of libvirtd virtualization backend
    section. It tries to validate options and combination of options that
    are specific for this virtualization backend.
    """

    VIRT_TYPE = 'libvirt'
    HYPERVISOR_ID = ('uuid', 'hostname')

    def __init__(self, section_name, wrapper, *args, **kwargs):
        super(LibvirtdConfigSection, self).__init__(section_name, wrapper, *args, **kwargs)
        # Note: no option is required for this virtualization backend. When no options
        # are specified, then virt-who will try to gather information from localhost

    def _validate_server(self, key):
        """
        Do validation of server option specific for this virtualization backend
        return: None or info-warning/error
        """
        # Server option is optional for libvirtd (libvirt hypervisor runs on the same
        # machine as virt-who)
        result = []
        if key not in self._values:
            self._values[key] = ''
        else:
            username = None
            url = self._values[key]
            if "//" not in url:
                url = "//" + url
            splitted_url = urllib.parse.urlsplit(url)

            hostname = splitted_url.hostname

            if hostname is None:
                hostname = ''

            if splitted_url.scheme:
                scheme = splitted_url.scheme
            else:
                result.append((
                    'info',
                    "Protocol is not specified in libvirt url, using qemu+ssh://"
                ))
                scheme = 'qemu+ssh'

            if 'username' in self._values:
                username = self._values['username']
            elif splitted_url.username:
                username = splitted_url.username

            if len(splitted_url.path) > 1:
                path = splitted_url.path
            else:
                result.append((
                    'info',
                    "Libvirt path is not specified in the url, using /system"
                ))
                path = '/system'

            self._values[key] = "%(scheme)s://%(username)s%(hostname)s%(path)s?no_tty=1" % {
                'username': ("%s@" % username) if username else '',
                'scheme': scheme,
                'hostname': hostname,
                'path': path
            }

        if len(result) == 0:
            return None
        else:
            return result

    def _validate_env(self, key):
        """
        Do validation of environment (env)
        :return: None or tuple with warning, when options are wrong
        """
        result = None
        sm_type = self._values['sm_type']
        if sm_type == 'sam' and 'server' in self._values and key not in self._values:
            result = (
                'error',
                "Option `%s` needs to be set in config: '%s'" % (key, self.name)
            )
        return result

    def _validate_owner(self, key):
        """
        Do validation of owner
        :return: None or tuple with warning, when options are wrong
        """
        result = None
        sm_type = self._values['sm_type']
        if sm_type == 'sam' and 'server' in self._values and key not in self._values:
            result = (
                'error',
                "Option `%s` needs to be set in config: '%s'" % (key, self.name)
            )
        return result

    def __validate_password(self, pass_key):
        """
        Validate password (encrypted or unecrypted) for libvirtd backend
        :return: None or warning/info, when options are wrong
        """
        result = None
        if pass_key not in self._values:
            return result
        else:
            if 'server' in self._values:
                server = self._values['server']
                if 'ssh://' in server or '://' not in server:
                    result = [(
                        'warn',
                        "Password authentication doesn't work with ssh transport on libvirt backend, "
                        "copy your public ssh key to the remote machine"
                    )]
            else:
                result = [(
                    'info',
                    "Options '%s' set in %s is useless for monitoring local libvirtd instance" %
                    (pass_key, self.name)
                )]
        return result

    def _validate_encrypted_password(self, pass_key):
        """
        Validate encrypted password
        :param pass_key: Key of encrypted password
        :return: None or tuple with warning, when options are wrong
        """
        if pass_key == 'encrypted_password':
            result = self.__validate_password(pass_key)
        else:
            result = super(LibvirtdConfigSection, self)._validate_encrypted_password(pass_key)
        return result

    def _validate_unencrypted_password(self, pass_key):
        """
        Validate unencrypted password
        :param pass_key: Key of unencrypted password
        :return: None or tuple with warning, when options are wrong
        """
        if pass_key == 'password':
            result = self.__validate_password(pass_key)
        else:
            result = super(LibvirtdConfigSection, self)._validate_unencrypted_password(pass_key)

        return result


class LibvirtdGuest(Guest):
    def __init__(self, domain):
        try:
            state = domain.state(0)[0]
        except AttributeError:
            # Some versions of libvirt doesn't have domain.state() method,
            # use first value from info instead
            state = domain.info()[0]

        super(LibvirtdGuest, self).__init__(
            uuid=domain.UUIDString(),
            virt_type=Libvirtd.CONFIG_TYPE,
            state=state)


class VirEventLoopThread(threading.Thread):
    def __init__(self, logger, *args, **kwargs):
        self._terminated = threading.Event()
        threading.Thread.__init__(self, *args, **kwargs)

    def run(self):
        while not self._terminated.is_set():
            libvirt.virEventRunDefaultImpl()

    def terminate(self):
        self._terminated.set()


def libvirt_cred_request(credentials, config):
    """ Callback function for requesting credentials from libvirt """
    for credential in credentials:
        if credential[0] == libvirt.VIR_CRED_AUTHNAME:
            credential[4] = config.get('username', None)
        elif credential[0] == libvirt.VIR_CRED_PASSPHRASE:
            credential[4] = config.get('password', None)
        else:
            return -1
    return 0


class Libvirtd(Virt):
    """ Class for interacting with libvirt. """
    CONFIG_TYPE = "libvirt"

    def __init__(self, logger, config, dest, terminate_event=None,
                 interval=None, oneshot=False, status=False, registerEvents=True):
        super(Libvirtd, self).__init__(logger, config, dest,
                                       terminate_event=terminate_event,
                                       interval=interval,
                                       oneshot=oneshot,
                                       status=status)
        self.changedCallback = None
        self.registerEvents = registerEvents
        self._host_capabilities_xml = None
        self._host_socket_count = None
        self._host_uuid = None
        self._host_name = None
        self.eventLoopThread = None
        libvirt.registerErrorHandler(lambda ctx, error: None, None)

    def getVersion(self):
        """
        The constants used to extract the version numbers were found in
        /lib64/python2.7/site-packages/libvirt.py
        """
        version_num = self.virt.getVersion()
        major = version_num / 1000000
        version_num -= major * 1000000

        minor = version_num / 1000
        version_num -= minor * 1000

        release = version_num
        return "%(major)s.%(minor)s.%(release)s" % {
            'major': major,
            'minor': minor,
            'release': release
        }

    def getHypervisorType(self):
        return self.virt.getType()

    def isHypervisor(self):
        return bool(self.config.get('server', None))

    def _createEventLoop(self):
        libvirt.virEventRegisterDefaultImpl()
        if self.eventLoopThread is not None and self.eventLoopThread.is_alive():
            self.eventLoopThread.terminate()
        self.eventLoopThread = VirEventLoopThread(self.logger, name="libvirtEventLoop")
        self.eventLoopThread.setDaemon(True)
        self.eventLoopThread.start()

    def _connect(self):
        url = self.config.get('server', "")
        self.logger.info("Using libvirt url: %s", url if url else '""')
        try:
            if self.config.get('password', None):
                auth = [[libvirt.VIR_CRED_AUTHNAME, libvirt.VIR_CRED_PASSPHRASE], libvirt_cred_request, self.config]
                v = libvirt.openAuth(url, auth, libvirt.VIR_CONNECT_RO)
            else:
                v = libvirt.openReadOnly(url)
        except libvirt.libvirtError as e:
            self.logger.exception("Error in libvirt backend")
            raise VirtError(str(e))
        v.domainEventRegister(self._callback, None)
        v.setKeepAlive(5, 3)
        return v

    def _disconnect(self):
        if self.virt is None:
            return
        try:
            self.virt.domainEventDeregister(self._callback)
            self.virt.close()
        except libvirt.libvirtError:
            pass
        self.virt = None

    def _run(self):
        self._createEventLoop()

        self.virt = None
        initial = True
        self.next_update = None

        while not self.is_terminated():
            if self.virt is None:
                self.virt = self._connect()

            if self.virt.isAlive() != 1:
                self._disconnect()
                self.virt = self._connect()

            if initial:
                report = self._get_report()
                self._send_data(data_to_send=report)
                initial = False
                self.next_update = time.time() + self.interval

            if self._oneshot:
                break

            time.sleep(1)
            if time.time() > self.next_update:
                report = self._get_report()
                self._send_data(data_to_send=report)
                self.next_update = time.time() + self.interval

        if self.eventLoopThread is not None and self.eventLoopThread.is_alive():
            self.eventLoopThread.terminate()
            self.eventLoopThread.join(1)
        self._disconnect()

    def _callback(self, *args, **kwargs):
        report = self._get_report()
        self._send_data(data_to_send=report)
        self.next_update = time.time() + self.interval

    def _get_report(self):
        if self.status:
            return StatusReport(self.config)
        if self.isHypervisor():
            return HostGuestAssociationReport(self.config, self._getHostGuestMapping())
        else:
            return DomainListReport(self.config, self._listDomains(), self._remote_host_id())

    def _lookupDomain(self, method, domain):
        '''
        Attempt to find the domain using given method.

        Returns None if the domain does not exist (it was probably just destroyed)
        '''
        try:
            return method(domain)
        except libvirt.libvirtError as e:
            if e.get_error_code() == libvirt.VIR_ERR_NO_DOMAIN:
                # Domain not found, most likely it was just destroyed
                return None
            else:
                # All other exceptions should be forwarded
                raise

    def _listDomains(self):
        domains = []
        try:
            # Active domains
            for domainID in self.virt.listDomainsID():
                domain = self._lookupDomain(self.virt.lookupByID, domainID)
                if domain is None:
                    # Domain not found, most likely it was just destroyed, ignoring
                    self.logger.debug("Lookup for domain by ID %s failed, probably it was just destroyed, ignoring" % domainID)
                    continue

                if domain.UUIDString() == "00000000-0000-0000-0000-000000000000":
                    # Don't send Domain-0 (zeroed uuid)
                    continue
                domains.append(LibvirtdGuest(domain))

            # Non active domains
            for domainName in self.virt.listDefinedDomains():
                domain = self._lookupDomain(self.virt.lookupByName, domainName)
                if domain is None:
                    # Domain not found, most likely it was just destroyed, ignoring
                    self.logger.debug("Lookup for domain by name '%s' failed, probably it was just destroyed, ignoring" % domainName)
                    continue

                domains.append(LibvirtdGuest(domain))
        except libvirt.libvirtError as e:
            self.virt.close()
            raise VirtError(str(e))
        self.logger.debug("Libvirt domains found: %s", ", ".join(guest.uuid for guest in domains))
        return domains

    @property
    def host_capabilities_xml(self):
        if self._host_capabilities_xml is None:
            self._host_capabilities_xml = ElementTree.fromstring(self.virt.getCapabilities())
        return self._host_capabilities_xml

    def _remote_host_id(self):
        if self._host_uuid is None and 'hypervisor_id' in self.config:
            if self.config['hypervisor_id'] == 'uuid':
                self._host_uuid = self.host_capabilities_xml.find('host/uuid').text
            elif self.config['hypervisor_id'] == 'hostname':
                self._host_uuid = self.virt.getHostname()
        return self._host_uuid

    def _remote_host_name(self):
        if self._host_name is None:
            try:
                self._host_name = self.virt.getHostname()
            except AttributeError:
                self._host_name = None
        return self._host_name

    def _remote_host_sockets(self):
        if self._host_socket_count is None:
            try:
                self._host_socket_count = self.host_capabilities_xml.find('host/cpu/topology').get('sockets')
            except AttributeError:
                self._host_socket_count = None
        return self._host_socket_count

    def _getHostGuestMapping(self):
        mapping = {'hypervisors': []}
        facts = {
            Hypervisor.CPU_SOCKET_FACT: self._remote_host_sockets(),
            Hypervisor.HYPERVISOR_TYPE_FACT: self.virt.getType(),
            Hypervisor.HYPERVISOR_VERSION_FACT: self.virt.getVersion(),
            Hypervisor.SYSTEM_UUID_FACT: self.host_capabilities_xml.find('host/uuid').text,
        }
        host = Hypervisor(hypervisorId=self._remote_host_id(),
                          guestIds=self._listDomains(),
                          name=self._remote_host_name(),
                          facts=facts)
        mapping['hypervisors'].append(host)
        return mapping



from __future__ import print_function

#
# Module for abstraction of all virtualization backends, part of virt-who
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

import os
import six
import sys
import tempfile
import ssl
import json
import shutil
import requests

from fake_server import FakeServer, FakeHandler

from rhsm import config as rhsm_config


class SamHandler(FakeHandler):
    """
    Handler of GET, POST and PUT method for FakeSam (candlepin server)
    """

    def do_GET(self):
        """
        Handler for GET method
        :return:
        """
        print("[FakeSam] GET", self.path)
        if self.path.startswith('/status'):
            content = {
                "result": "ok",
                "managerCapabilities": [""]
            }
            self.write_json(content)
        else:
            content = {"result": "ok"}
            self.write_json(content)

    def do_POST(self):
        """
        Handler for POST method
        """
        print("[FakeSam] POST", self.path, self.server.code)
        if self.server.code:
            self.write_json({}, status_code=self.server.code, headers={"Retry-After": "60"})
        elif self.path.startswith('/hypervisors') or self.path.startswith('//hypervisors'):
            size = int(self.headers["Content-Length"])
            incoming = self.rfile.read(size)
            if isinstance(incoming, six.binary_type):
                incoming = incoming.decode('utf-8')
            data = json.loads(incoming)
            print("[FakeSam] putting in the queue:", data)
            self.server.queue.put(data)
            content = {
                "failedUpdate": [],
                "updated": [],
                "created": [],
            }
            self.write_json(content, status_code=requests.codes.ok)

    def do_PUT(self):
        """
        Handler for PUT method. This method is just ignored
        :return:
        """
        print("[FakeSam] PUT", self.path)


class FakeSam(FakeServer):
    """
    Fake candlepin server used for testing of virt-who
    """
    def __init__(self, queue, port=None, code=None, host='localhost'):
        """
        Initialization of fake candlepin server
        :param queue: inter-process queue used for testing
        :param port: port, where server is listening on
        :param code: (optional) the code that server returns to client
        :param host: the name that is used for host
        """
        super(FakeSam, self).__init__(SamHandler, port=port, host=host)
        self.server.code = code
        base = os.path.dirname(os.path.abspath(__file__))
        certfile = os.path.join(base, 'cert.pem')
        keyfile = os.path.join(base, 'key.pem')
        if not os.access(certfile, os.R_OK):
            raise OSError("No such file %s" % certfile)
        if not os.access(keyfile, os.R_OK):
            raise OSError("No such file %s" % keyfile)
        self.server.socket = ssl.wrap_socket(self.server.socket, certfile=certfile, keyfile=keyfile, server_side=True)

        self.tempdir = tempfile.mkdtemp()
        config_name = os.path.join(self.tempdir, 'rhsm.conf')
        with open(config_name, 'w') as f:
            f.write("""
[server]
hostname = {host}
prefix = /
port = {port}
insecure = 1
proxy_hostname =

[rhsm]
consumerCertDir = {certdir}
""".format(host=self.host, port=self.port, certdir=base))

        rhsm_config.DEFAULT_CONFIG_PATH = config_name

        self.server.sam = self
        self.server.queue = queue

    def terminate(self):
        shutil.rmtree(self.tempdir)
        super(FakeSam, self).terminate()


if __name__ == '__main__':
    if len(sys.argv) >= 2:
        code = int(sys.argv[1])
    else:
        code = None
    from six.moves.queue import Queue
    q = Queue()
    f = FakeSam(q, port=8443, code=code, host='0.0.0.0')
    f.run()
